/*
  Run this Dfa on `input' and return the resulting state
 */
function Dfa_runState(input) {
  input = input.toString();
  var table = this.table;
  if ((table == null) || (input == null) || ((table.length == 1) && !table[0][0])) {
    return -1;
  }
  var current_state = 0;
  for (var i = 0; (i < input.length) && (current_state >= 0); i++) {
    var c = input.charAt(i);
    var state = current_state;
    current_state = -1;
    for (var j = 1; j < table[state].length; j += 3) {
      if ((table[state][j] <= c) && (c <= table[state][j+1])) {
        current_state = table[state][j+2];
        break;
      }
    }
  }
  return current_state;
}

/*
  run this Dfa on input and return whether the dfa accepts, rejects,
  or crashes
 */
function Dfa_run(input) {
  var state = this.runState(""+input);
  if (state < 0) {
    return PowerForms.CRASH;
  } else if (!this.table[state][0]) {
    return PowerForms.REJECT;
  } else {
    return PowerForms.ACCEPT;
  }
}

/*
  check if one string in values is accepted by this Dfa
 */
function Dfa_match(values) {
  if (typeof(values) == "string") {
    return this.run(values) == PowerForms.ACCEPT;
  } else {
    for (var i = 0; i < values.length; i++) {
      var value = values[i];
      if (this.run(value) == PowerForms.ACCEPT) {
	return true;
      }
    }
    return false;
  }
}

/*
  check if all strings in values are accepted by this Dfa
 */
function Dfa_matchAll(inputs) {
  if (typeof(values) == "string") {
    return this.run(values) == PowerForms.ACCEPT;
  } else {
    for (var i = 0; i < values.length; i++) {
      var value = values[i];
      if (this.run(value) != PowerForms.ACCEPT) {
	return false;
      }
    }
    return true;
  }
}

function Dfa_accept(input) {
  return this.run(input) == PowerForms.ACCEPT;
}

/*
  return string representation of this Dfa
 */
function Dfa_toString() {
  var table = this.table;
  if (table == null) {
    return "[The empty dfa]";
  }
  var str = "[Dfa: "+table.length+" state"+(table.length == 1 ? "" : "s");
  for (var i = 0; i < table.length; i++) {
    str += "\n"+i+": "+table[i];
  }
  return str+"]";
}

/*
  if prefix then complete with a unique prefix possibly leading to a
  rejecting state
 */
function Dfa_completeUnique(input, prefix) {
  var state = this.runState(input);
  var table = this.table;
  var buffer = "";
  while ((state > -1) &&
         (!table[state][0]) &&
         ((table[state].length == 4) && (table[state][1] == table[state][2]))) {
    buffer += table[state][1];
    state = table[state][3];
  }
  if ((state < 0) || (!prefix && !table[state][0])) {
    return null;
  }
  return buffer;
}

function _completeAllSearch(dfa, state, prefix, visited) {
  var completions = new Array();
  if (!visited[state]) {
    if (dfa.table[state][0]) {
      if (completions.push) {
	completions.push(prefix);
      } else {
	completions[completions.length] = prefix;
      }
    } else {
      visited[state] = true;
      var transitions = dfa.table[state];
      var new_visited = new Array();
      for (var i = 1; i < transitions.length; i += 3) {
        new_visited = visited.concat();
        var c = _completeAllSearch(dfa, transitions[i+2], prefix+transitions[i], new_visited);
	if (completions.concat) {
	  completions = completions.concat(c);
	} else {
	  for (var j = 0; j < c.length; j++) {
	    completions[completions.length] = c[j];
	  }
	}
      }
    }
  }
  return completions;
}

function Dfa_completeAll(input) {
  if ((input == null) || ((this.table.length == 1) && (this.table[0][0] == FALSE))) {
    return null;
  }
  var visited = new Array(this.table.length);
  for (var i = 0; i < visited.length; i++) {
    visited[i] = false;
  }
  var current_state = 0;
  for (var i = 0; (i < input.length) && (current_state > -1); i++) {
    var c = input.charAt(i);
    var state = current_state;
    visited[state] = true;
    current_state = -1;
    for (var j = 1; j < this.table[state].length; j += 3) {
      if ((this.table[state][j] <= c) && (c <= this.table[state][j+1])) {
        current_state = this.table[state][j+2];
        break;
      }
    }
  }
  var state = current_state;
  if ((state > -1) && !this.table[state][0]) {
    return _completeAllSearch(this, state, "", visited);
  }
  return null;
}

function Dfa_completeFirst(input) {
  var completions = this.completeAll(input);
  return completions == null ? null : completions[0];
}

/*
  Dfa constructor
 */
function Dfa(table) {
  this.table = table;
  this.runState = Dfa_runState;
  this.run = Dfa_run;
  this.match = Dfa_match;
  this.matchAll = Dfa_matchAll;
  this.completeUnique = Dfa_completeUnique;
  this.completeAll = Dfa_completeAll;
  this.completeFirst = Dfa_completeFirst;
  this.accept = Dfa_accept; //function(s) { return this.match(s); }
  this.toString = Dfa_toString;
}

/* The empty dfa */
Dfa.EMPTY = new Dfa(null);
Dfa.EMPTY.filter = function(s) { return false; }

/* Dfa accepting a single character */
Dfa.ANYCHAR = new Dfa(new Array(new Array(false, '\0', '\uFFFF', 1), new Array(true)));
Dfa.ANYCHAR.filter = function(s) { return Dfa.ANYTHING.match(s); }

/* Dfa accepting any string */
Dfa.ANYTHING = new Dfa(new Array(new Array(true, '\0', '\uFFFF', 0)));
Dfa.ANYTHING.filter = function(s) { return true; }
