/*
   Compute 3-state and.
*/
PowerForms.and = function() {
  switch (arguments.length) {
  case 0:
    return PowerForms.NO;
  case 1:
    return arguments[0];
  default:
    var result = PowerForms.YES;
    for (var i = 0; i < arguments.length; i++) {
      if (arguments[i] == PowerForms.NO) {
	return PowerForms.NO;
      } else if (arguments[i] == PowerForms.MAYBE) {
	result = PowerForms.MAYBE;
      }
    }
    return result;
  }
}

PowerForms.negate = function(s) {
  switch (s) {
  case PowerForms.NO:
    return PowerForms.YES;
  case PowerForms.MAYBE:
    return PowerForms.MAYBE;
  default:
    return PowerForms.NO;
  }
}

/*
  Compute 3-state not.
*/
PowerForms.not = function() {
  switch (arguments.length) {
  case 0:
    return PowerForms.NO;
  case 1:
    return PowerForms.negate(arguments[0]);
  default:
    var result = PowerForms.YES;
    for (var i = 0; i < arguments.length; i++) {
      if (PowerForms.negate(arguments[i]) == PowerForms.NO) {
	return PowerForms.NO;
      } else if (PowerForms.negate(arguments[i]) == PowerForms.MAYBE) {
	result = PowerForms.MAYBE;
      }
    }
    return result;
  }
}

/*
   Compute 3-state or.
*/
PowerForms.or = function() {
  var result = PowerForms.NO;
  for (var i = 0; i < arguments.length; i++) {
    if (arguments[i] == PowerForms.YES) {
      return PowerForms.YES;
    } else if (arguments[i] == PowerForms.MAYBE) {
      result = PowerForms.MAYBE;
    }
  }
  return result;
}

/**
   Compare 2 strings for equality
   @param s of type string
   @param t of type string
   @return PowerForms.YES if the strings are equal,
           PowerForms.MAYBE if the first string is a prefix of the other, and
	   PowerForms.NO otherwise.
 */
PowerForms.equals = function(s, t) {
  /* confirm("equal(`"+s+"', `"+t+"') ..."); */
  if (s == null) {
    return t == null;
  } else if (s == t) {
    return PowerForms.YES;
  } else {
    s = s.toString();
    t = t.toString();
    var i;
    for (i = 0; (i < s.length) && (i < t.length); i++) {
      if (s.charAt(i) != t.charAt(i)) {
	return PowerForms.NO;
      }
    }
    return (i < t.length ? PowerForms.MAYBE : PowerForms.NO);
  }
}

/* arguments: ``form, field, form, field'' or
              ``form, field, value'' */
PowerForms.equal = function() {
  var length = arguments.length;
  if (length%2 == 1) {
    length--;
    var value = arguments[length];
    var v = PowerForms.getValues(arguments[0], arguments[1]);
    /* PowerForms.DEBUG("equal: value: "+value+"; values: "+v); */
    if (v.member(value)) {
      return PowerForms.YES;
    }
    var elements = v.getDynamicElements();
    // for (var i = 0; i < elements.length; i++) {
    for (var i in elements) {
      if (value.indexOf(elements[i]) == 0) {
	return PowerForms.MAYBE;
      }
    }
    return PowerForms.NO;
  } else {
    var v0 = PowerForms.getValues(arguments[0], arguments[1]);
    var v1 = PowerForms.getValues(arguments[2], arguments[3]);
    var elements0 = v0.getElements();
    var elements1 = v1.getElements();

    /* window.status = (v0+"; "+v1); */

    // FIXME
    for (var i = 0; i < elements0.length; i++) {
      if (v1.member(elements0[i])) {
	return PowerForms.YES;
      }
    }

    var result = PowerForms.NO;
    for (var i = 0; i < elements0.length; i++) {
      var e0 = elements0[i].toString();
      for (var j = 0; j < elements1.length; j++) {
	var e1 = elements1[j].toString();
	if ((v1.getType(e1) == PowerForms.DYNAMIC) && (e0.indexOf(e1) == 0)
	    || (v0.getType(e0) == PowerForms.DYNAMIC) && (e1.indexOf(e0) == 0)) {
	  result = PowerForms.MAYBE;
	}
      }
    }
    return result;
  }
  // FIXME: Prefix-lighed (gult lys)???
}

/* arguments: ``value, value'' or
              ``value, form, field'' */
PowerForms.equalValue = function() {
  var value = arguments[0];
  if (arguments.length == 2) {
    var v = arguments[1];
    return PowerForms.equals(value, v);
  } else {
    var result = PowerForms.NO;
    var v = PowerForms.getValues(arguments[1], arguments[2]);
    var elements = v.getElements();
    for (var i = 0; i < elements.length; i++) {
      var s = PowerForms.equals(value, elements[i]);
      if (s == PowerForms.YES) {
	return PowerForms.YES;
      } else if (s == PowerForms.MAYBE) {
	result = PowerForms.MAYBE;
	/* confirm(value+" might equal "+elements[i]); */
      }
    }
    return result;
  }
}

/* arguments: ``value, value, prefex'' or
              ``value, form, field, prefix'' */
PowerForms.equalCompleteUnique = function() {
  if (arguments.length == 3) {
    var value = arguments[0].toString();
    var v = arguments[1].toString();
    if (PowerForms.equals(value, v) == PowerForms.MAYBE) {
      return v.substring(value);
    }
  } else {
    var value = arguments[0];
    var v = PowerForms.getValues(arguments[1], arguments[2]);
    var elements = v.getElements();
    var completions = new Array();
    for (var i = 0; i < elements.length; i++) {
      var s = PowerForms.equals(value, elements[i]);
      if (s == PowerForms.YES) {
	return null;
      } else if (s == PowerForms.MAYBE) {
	completions[completions.length] = elements[i].toString().substring(value.length);
      }
    }
    if (completions.length == 1) {
      return completions[0];
    } else if (prefix) {
      // return common prefix of all completions excluding the empty string
    }
  }
  return null;
}

/* arguments: ``value, value'' or
              ``value, form, field'' */
PowerForms.equalCompleteFirst = function() {
  if (arguments.length == 2) {
    return PowerForms.equalCompleteUnique(arguments[0], arguments[1], false);
  } else {
    var value = arguments[0];
    var v = PowerForms.getValues(arguments[1], arguments[2]);
    var elements = v.getElements();
    var first = null
    for (var i = 0; i < elements.length; i++) {
      var s = PowerForms.equals(value, elements[i]);
      if (s == PowerForms.YES) {
	return null;
      } else if (s == PowerForms.MAYBE) {
	var completion = elements[i].toString().substring(value.length);
	if ((first == null) || (completion < first )) {
	  first = completion;
	}
      }
    }
    return first;
  }
}


PowerForms.count = function(form, name, min, max) {
  var size = PowerForms.getValues(form, name).size();
  if ((min <= size) && (size <= max)) {
    return PowerForms.YES;
  }
  return PowerForms.NO;
}

PowerForms.match = function(form, name, dfa) {
  var elements = PowerForms.getValues(form, name).getElements();
  var result = PowerForms.NO;
  for (e in elements) {
    var r = dfa.run(elements[e]);
    /* window.status = "e: "+elements[e]+" ("+elements.length+"); r: "+r; */
    if (r == PowerForms.ACCEPT) {
      return PowerForms.YES;
    }
    if (r == PowerForms.REJECT) {
      result = PowerForms.MAYBE;
    }
  }
  return result;
}

PowerForms.isInteger = function(s) {
  return s == parseInt(s).toString();
}

PowerForms.isIntegerPrefix = function(s) {
  return (s == "") || (s == "-") || PowerForms.isInteger(s);
}

/**
   Compare 2 strings.
   @param a of type string
   @param b of type string
   @param type the comparison of comparison to be performed. Allowed
               values are PowerForms.STRING and PowerForms.INTEGER.
   @return PowerForms.YES if a is less than b
           PowerForms.MAYBE if a is a prefix of a string that is less than b
	   PowerForms.NO otherwise.
 */
PowerForms.lessThanEval = function(a, b, type) {
  if (type == PowerForms.STRING) {
    a = a.toString();
    b = b.toString();
    if (a < b) {
      return PowerForms.YES;
    } else if ((a.indexOf(b) == 0) || (b.indexOf(a) == 0)) {
      return PowerForms.MAYBE;
    } else {
      return PowerForms.NO;
    }
  } else if (type == PowerForms.INTEGER) {
    var i = parseInt(a);
    var j = parseInt(b);
    if (!isNaN(i) && !isNaN(j)) {
      if (i < j) {
	return PowerForms.YES;
      } else if ((0 <= i) && (j < 0)) {
	return PowerForms.NO;
      } else {
	return PowerForms.MAYBE;
      }
    } else if (!isNaN(i) && PowerForms.isIntegerPrefix(b)) {
      if ((i >= 0) && (b == "-")) {
	return PowerForms.NO;
      } else {
	return PowerForms.MAYBE;
      }
    } else if (PowerForms.isIntegerPrefix(a) && !isNaN(j)) {
      return PowerForms.MAYBE;
    } else if (PowerForms.isIntegerPrefix(a) && PowerForms.isIntegerPrefix(b)) {
      return PowerForms.MAYBE;
    } else {
      return PowerForms.NO;
    }
  }
  /* alert("Unhandled case: lessthan("+a+", "+b+", "+type+") ..."); */
  return PowerForms.NO;
}

/* arguments: ``form, field, form, field, type'' or
              ``form, field, value, type'' */
PowerForms.lessThan = function() {
  var length = arguments.length;
  var type = arguments[--length];
  if (length%2 == 1) {
    length--;
    var value = arguments[length];
    var v = PowerForms.getValues(arguments[0], arguments[1]);
    var elements = v.getElements();
    for (var i = 0; i < elements.length; i++) {
      if (PowerForms.lessThanEval(elements[i], value, type) == PowerForms.YES) {
	return PowerForms.YES;
      }
    }
    elements = v.getDynamicElements();
    for (var i = 0; i < elements.length; i++) {
      if (PowerForms.lessThanEval(elements[i], value, type) == PowerForms.MAYBE) {
	return PowerForms.MAYBE;
      }
    }
    return PowerForms.NO;
  } else {
    var v0 = PowerForms.getValues(arguments[0], arguments[1]);
    var v1 = PowerForms.getValues(arguments[2], arguments[3]);
    var elements0 = v0.getElements();
    var elements1 = v1.getElements();

    for (var i0 = 0; i0 < elements0.length; i0++) {
      for (var i1 = 0; i1 < elements0.length; i1++) {
	if (PowerForms.lessThanEval(elements0[i0], elements1[i1], type) == PowerForms.YES) {
	  return PowerForms.YES;
	}
      }
    }

    for (var i0 = 0; i0 < elements0.length; i0++) {
      var e0 = elements0[i0].toString();
      for (var i1 = 0; i1 < elements1.length; i1++) {
	var e1 = elements1[i1].toString();
	if (((v1.getType(e1) == PowerForms.DYNAMIC) || (v0.getType(e0) == PowerForms.DYNAMIC))
	    && (PowerForms.lessThanEval(e0, e1, type) == PowerForms.MAYBE)) {
	  return PowerForms.MAYBE;
	}
      }
    }
    return PowerForms.NO;
  }
  // FIXME: Prefix-lighed (gult lys)???
}
