/**
 * AutoComplete.java
 *
 *
 * Created: Tue Apr 16 19:17:05 2002
 *
 * @author <a href="mailto:ricky@brics.dk">Mikkel Ricky</a>
 * @version
 */

package dk.brics.powerforms;

import java.io.*;
import java.net.*;

class AutoComplete extends PwfElement {
  final static int UNIQUE = 0;
  final static int FIRST = 1;
  final static int PREFIX = 2;
  final static int AGGRESSIVE = 3;

  ConstraintApply constraint;
  HtmlInput input;
  URL imageurl, naurl;
  int type = UNIQUE;
  int index = 0;

  public AutoComplete(Attributes attributes) {
    init("autocomplete", attributes);
  }

  String getScriptName() {
    return "pwf_autocomplete_"+index;
  }

  void check() throws TransformException {
    String constraintid = attributes.requireValue("constraint");
    if (constraintid != null) {
      constraint = getPowerForms().getConstraint(constraintid);
      if (constraint == null) {
	reportError("no constraint with id `"+constraintid+"' defined");
      } else {
	// FIXME: check that constraint is a field-constraint ...
	constraint.addAutoComplete(this);
      }
    }
    try {
      imageurl = new URL(getPowerForms().getMiscUrl(), attributes.getValue("image", "autocomplete.gif"));
      naurl = new URL(getPowerForms().getMiscUrl(), attributes.getValue("na", "nocompletion.gif"));
    } catch (MalformedURLException e) {
      throw new TransformException(e);
    }
    String t = attributes.getValue("type");
    if (t == null) {
      t = "unique";
    }
    if (t.equals("unique")) {
      type = UNIQUE;
    } else if (t.equals("first")) {
      type = FIRST;
    } else if (t.equals("prefix")) {
      type = PREFIX;
    } else if (t.equals("aggressive")) {
      type = AGGRESSIVE;
    } else {
      reportError("unknown type for completion: `"+t+"'");
    }
  }

  private boolean defined = false;
  private boolean initialized = false;

  void scriptDefine(HtmlDocument document, PrintWriter out) {
    if (!defined) {
      if ((constraint != null) || ((input != null) && input.hasConstraints())) {
	out.println("var "+getScriptName()+" = null;");
      }
      defined = true;
    }
  }

  void scriptInit(HtmlDocument document, PrintWriter out) {
    if (!initialized && defined) {
      String updateFunction = input.getDocument().getUpdateFunctionName(input.form.id, input.name);
      out.println("  "+getScriptName()+" = new AutoComplete('"+id+"'"
		  +", '"+input.id+"'"
		  +", "+updateFunction
		  +", '"+imageurl+"'"
		  +", '"+naurl+"')");
      initialized = true;
    }
  }

  void code(PrintWriter out) {
    if (initialized) {
      out.print("<a onmouseover='window.status = \"\";' id='"+id+"-anchor' name='"+id+"-anchor' href='javascript:"+getScriptName()+".complete();'>"
		+"<img border='0' id='"+id+"' name='"+id+"' src='"+naurl+"' alt='' />"
		+"</a>");
    }
  }

  AutoComplete copy(int index, HtmlInput input) {
    AutoComplete a = new AutoComplete(new Attributes(getPowerForms()));
    a.attributes.addID();
    a.constraint = constraint;
    a.index = index;
    a.input = input;
    a.imageurl = imageurl;
    a.naurl = naurl;
    a.type = type;
    return a;
  }

  public String toString() {
    CharArrayWriter chars = new CharArrayWriter();
    PrintWriter out = new PrintWriter(chars);
    out.println("AutoComplete");
    out.println("  Constraint: "+attributes.getValue("constraint"));
    out.println("  Image-url: "+imageurl);
    out.println("  Na-url: "+naurl);
    out.println("  Type: "+type);
    out.close();
    return chars.toString();
  }
}
