/*
 * PowerForms
 * Copyright (C) 1999-2002 BRICS.
 *
 * Author: Mikkel Ricky <ricky@brics.dk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the  Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * Home page: http://www.brics.dk/bigwig/powerforms/
 */

package dk.brics.powerforms;

import dk.brics.automaton.*;
import java.io.*;
import java.net.*;
import java.util.*;

public class HtmlDocument {
  PowerForms powerforms = null;
  HtmlContent content;
  Map elements;
  List forms;
  ListSet constraints = null;
  List constraintClasses = null;
  private Map formConstraints = null;
  Map fieldClasses = null;
  StringPairSet fields;
  boolean generate_script_code = true;
  boolean generate_noscript = true;
  boolean insert_logo = true;
  boolean initCode = false;
  boolean selectCode = true;
  boolean statusCode = false;
  boolean cssCode = false;
  private boolean initialization = false;

  private int form_number = 0;
  private int status_number = 0;
  private int autocomplete_number = 0;
  private int applet_number = 0;

  int getFormNumber() {
    return form_number++;
  }

  boolean isLastForm(int number) {
    return number == form_number-1;
  }

  int getStatusNumber() {
    return status_number++;
  }

  int getAppletNumber() {
    return applet_number++;
  }

  HtmlDocument(PowerForms powerforms) {
    elements = new HashMap();
    forms = new ArrayList();
    setPowerForms(powerforms);
  }

  void setPowerForms(PowerForms powerforms) {
    this.powerforms = powerforms;
  }

  PowerForms getPowerForms() {
    return powerforms;
  }

  void addElement(HtmlElement element) {
    if (element.id != null) {
      HtmlElement e = getElement(element.id);
      if (e != null) {
	/* element.reportError("an element with id `"+element.id+"' already declared ("+e.getLocation()+")"); */
      } else {
	elements.put(element.id, element);
      }
    }
  }

  HtmlElement getElement(String id) {
    return (HtmlElement)elements.get(id);
  }

  void addForm(HtmlForm form) {
    forms.add(form);
  }

  boolean hasConstraints(HtmlForm form) {
    return formConstraints.containsKey(form);
  }

  ListSet getConstraints(HtmlForm form) {
    ListSet s = (ListSet)formConstraints.get(form);
    if (s == null) {
      return null;
    } else {
      ListSet sorted = new ListSet();
      for (Iterator i = s.iterator(); i.hasNext(); ) {
	ConstraintApply c = (ConstraintApply)i.next();
	if (c.fieldName != null) {
	  sorted.add(c);
	}
      }
      for (Iterator i = s.iterator(); i.hasNext(); ) {
	ConstraintApply c = (ConstraintApply)i.next();
	if (c.fieldName == null) {
	  sorted.add(c);
	}
      }
      return sorted;
    }
  }

  /** Get form by id. If <var>id</var> is <code>null</code> return first form in
      this document.
   */
  HtmlForm getForm(String id) {
    if (id == null) {
      return forms.size() > 0 ? (HtmlForm)forms.get(0) : null;
    } else {
      for (int i = 0; i < forms.size(); i++) {
	HtmlForm form = (HtmlForm)forms.get(i);
	if (id.equals(form.id)) {
	  return form;
	} else if ((form.attributes != null) && id.equals(form.attributes.getValue("name"))) {
	  return form;
	}
      }
      return null;
    }
  }

  HtmlStatus getDefaultStatus(HtmlField field) {
    if ((getPowerForms().getConstraints(field.form.id, field.name) == null)
	|| ((field.form.index > 0) && (field.form.id == null))) {
      return null;
    } else {
      Attributes attributes = getPowerForms().getDefaultStatusAttributes();
      if (attributes != null) {
	if (field.id == null) {
	  field.attributes.addID();
	  addElement(field);
	}
	HtmlStatus status = null;
	attributes.add("default-status", "default-status");
	attributes.add("field", field.id);
	if (field.form != null) {
	  attributes.add("form", field.form.id);
	}
	statusCode = true;
	if ("css".equals(attributes.getValue("type"))) {
	  cssCode = true;
	}
	status = new HtmlStatus(attributes);
	attributes.addID();
	status.setDocument(this);
	try {
	  status.check();
	} catch (TransformException e) {}
	DEBUG("  => "+status);
	return status;
      }
    }
    return null;
  }

  AutoComplete getAutoComplete(HtmlInput input) {
    AutoComplete a = getPowerForms().getAutoComplete(input.form.id, input.name);
    if (a != null) {
      a = a.copy(autocomplete_number++, input);
    }
    return a;
  }

  void check() throws TransformException{
    if (content != null) {
      content.check();
    }
    compute();
  }

  private void addFormConstraint(StringPair p) {
    DEBUG("  "+p);
    ListSet s0 = getPowerForms().getDependencies(p);
    ListSet s1 = getPowerForms().getConstraints(p);
    DEBUG("    affects: "+s0);
    DEBUG("    constrained by: "+s1);
    if (s0 != null) {
      constraints.addAll(s0);
    }
    if (s1 != null) {
      HtmlForm form = getForm(p.first);
      ListSet s = (ListSet)formConstraints.get(form); //getConstraints(form);
      if (s == null) {
	s = new ListSet();
	formConstraints.put(form, s);
      }
      s.addAll(s1);
      constraints.addAll(s1);
    }
    if ((s0 != null) || (s1 != null)) {
      ListSet s = new ListSet();
      if (s0 != null) {
	s.addAll(s0);
      }
      if (s1 != null) {
	s.addAll(s1);
      }
      DEBUG("      s: "+s);
      int index = constraintClasses.indexOf(s);
      if (index < 0) {
	index = constraintClasses.size();
	constraintClasses.add(s);
      }
      fieldClasses.put(p, new Integer(index));
    }
    if (getPowerForms().getInitialization(p) != null) {
      initialization = true;
    }
  }

  private void compute() {
    fields = new StringPairSet();
    constraints = new ListSet();
    formConstraints = new HashMap();
    constraintClasses = new ArrayList();
    fieldClasses = new HashMap();
    for (int i = 0; i < forms.size(); i++) {
      HtmlForm form = (HtmlForm)forms.get(i);
      for (int j = 0; j < form.fields.size(); j++) {
	HtmlFieldList fieldList = (HtmlFieldList)form.fields.get(j);
	fields.add(form.id, fieldList.name);
      }
      if (form.fields.size() > 0) {
	addFormConstraint(new StringPair(form.id, null));
      }
    }
    DEBUG("compute; fields: "+fields);
    for (Iterator i = fields.iterator(); i.hasNext(); ) {
      StringPair p = (StringPair)i.next();
      addFormConstraint(p);
    }
    DEBUG("  constraints: "+constraints);
    DEBUG("  formConstraints: "+formConstraints);
  }

  private void includeLibraries(PrintWriter out) {
    URL miscUrl = getPowerForms().getMiscUrl();
    try {
      URL url = new URL(miscUrl, "PowerForms.js");
      out.println("<script type='text/javascript' src='"+url+"'></script>");
      url = new URL(miscUrl, "Select.js");
      out.println("<script type='text/javascript' src='"+url+"'></script>");
      url = new URL(miscUrl, "Expression.js");
      out.println("<script type='text/javascript' src='"+url+"'></script>");
      url = new URL(miscUrl, "AutoComplete.js");
      out.println("<script type='text/javascript' src='"+url+"'></script>");
      if (getPowerForms().getRegExps().size() > 0) {
	url = new URL(miscUrl, "Dfa.js");
	out.println("<script type='text/javascript' src='"+url+"'></script>");
      }
      if (true || statusCode) {
	url = new URL(miscUrl, "Status.js");
	out.println("<script type='text/javascript' src='"+url+"'></script>");
	if (true || cssCode) {
	  url = new URL(miscUrl, "PowerForms.css");
	  out.println("<link href='"+url+"' rel='stylesheet' type='text/css' />");
	  /*
	  out.println("<style type='text/css'>");
	  out.println("  input.pwfRed { background-color: #faa; }");
	  out.println("  input.pwfYellow { background-color: #ffa; }");
	  out.println("  input.pwfGreen {}");
	  out.println("  input.pwfNa {}");
	  out.println("</style>");
	  */
	}
      }
    } catch (MalformedURLException e) {}
  }

  private void defineDfas(PrintWriter out) {
    for (Iterator i = getPowerForms().getRegExps().iterator(); i.hasNext(); ) {
      RegExp regexp = (RegExp)i.next();
      regexp.scriptDefine(this, out);
    }
  }

  private void defineConstraints(PrintWriter out) {
    for (Iterator i = constraints.iterator(); i.hasNext(); ) {
      Constraint constraint = (Constraint)i.next();
      constraint.scriptDefine(this, out);
    }
  }

  private static final String updateFunctionPrefix = "pwf_update";

  String getUpdateFunctionName() {
    return updateFunctionPrefix;
  }

  String getUpdateFunctionName(String form, String field) {
    Integer i = (Integer)fieldClasses.get(new StringPair(form, field));
    return i == null ? null : updateFunctionPrefix+"_"+i.intValue();
  }

  String getEventName() {
    return "pwf_event";
  }

  private void defineUpdateFunctions(PrintWriter out) {
    if (constraintClasses.size() > 0) {
      out.println();
      out.println("function "+updateFunctionPrefix+"(functions) {");
      out.println("  var accept = true;");
      out.println("  var change = false;");
      out.println("  var msgs = new Array(functions.length);");
      out.println("  do {");
      out.println("    accept = true;");
      out.println("    change = false;");
      out.println("    for (var i = 0; i != functions.length; i++) {");
      out.println("      var info = functions[i]();");
      out.println("      if (info[1]) {");
      out.println("        change = true;");
      out.println("      }");
      out.println("      if (info[0]) {");
      out.println("        msgs[i] = null;");
      out.println("      } else {");
      out.println("        accept = false;");
      out.println("        msgs[i] = info[2];");
      out.println("      }");
      out.println("    }");
      out.println("  } while (change);");
      out.println("  var messages = new Array();");
      out.println("  for (var i = 0; i != msgs.length; i++) {");
      out.println("    if (msgs[i] != null) {");
      out.println("      for (var j = 0; j != msgs[i].length; j++) {");
      out.println("        PowerForms.addMessage(messages, msgs[i][j]);");
      out.println("      }");
      out.println("    }");
      out.println("  }");
      out.println("  return Array(messages.length == 0, change, messages);");
      out.println("}");

      for (int i = 0; i < constraintClasses.size(); i++) {
	ListSet set = (ListSet)constraintClasses.get(i);
	out.println();
	out.println("function "+updateFunctionPrefix+"_"+i+"(e) {");
	out.println("  "+getEventName()+" = e;");
	out.print("  return "+updateFunctionPrefix+"(new Array(");
	for (Iterator j = set.iterator(); j.hasNext(); ) {
	  ConstraintApply constraint = (ConstraintApply)j.next();
	  out.print(constraint.getScriptName());
	  if (j.hasNext()) {
	    out.print(", ");
	  }
	}
	out.println("));");
	out.println("}");
      }
    }
  }

  void defineSubmitFunctions(PrintWriter out) {
    for (int i = 0; i < forms.size(); i++) {
      HtmlForm form = (HtmlForm)forms.get(i);
      form.scriptDefine(this, out);
    }
  }

  String getInitFunctionName() {
    if (initialization || (constraintClasses.size() > 0)) {
      return "pwf_init";
    }
    return null;
  }

  void defineInitFunctions(PrintWriter out) {
    if (initialization || (constraintClasses.size() > 0)) {
      out.println("function "+getInitFunctionName()+"() {");
      out.println("  var element;");
      content.scriptInit(this, out);
      // FIXME
      for (int i = 0; i < forms.size(); i++) {
	HtmlForm form = (HtmlForm)forms.get(i);
	String updateFunctionName = form.getUpdateFunctionName();
	if (updateFunctionName != null) {
	  out.println("  "+updateFunctionName+"();");
	}
      }
      out.println("  if (!document.getElementById) {");
      out.println("    document.getElementById = PowerForms.getElementById;");
      out.println("  }");
      out.println("}");
      initCode = true;
    }
  }

  private boolean scripted = false;

  void script(PrintWriter out) {
    if (!scripted && generate_script_code) {
      /* compute(); */
      boolean script = (initialization || (constraintClasses.size() > 0));
      if (!script) {
	for (Iterator i = forms.iterator(); i.hasNext(); ) {
	  HtmlForm form = (HtmlForm)i.next();
	  if (form.hasSubmitElement) {
	    script = true;
	    break;
	  }
	}
      }
      if (script) {
	out.println("<!-- PowerForms script code -->");
	includeLibraries(out);
	out.println("<script type='text/javascript'><!--");
	out.println("var "+getEventName()+" = null;");
	defineDfas(out);
	defineConstraints(out);
	defineUpdateFunctions(out);
	defineSubmitFunctions(out);
	defineInitFunctions(out);
	out.println("//--></script>");
	out.println("<!-- PowerForms script code -->");
      }
    }
    scripted = true;
  }

  void code(PrintWriter out) {
    if (content != null) {
      if (docType != null) {
	out.println(docType);
      }
      /* out.println("<?xml version='1.0' encoding='iso-8859-1'?>"); */
      out.println();
      content.code(out);
    }
  }

  /**
     @param values a <code>Map</code> from <code>String</code>s to
     <code>Collection</code>s
     @param handler an <code>ErrorHandler</code> to receive
     notifications of validation errors
     @return <var>true</var> iff validation succeds
   */
  public boolean validate(Map values, ErrorHandler handler, boolean jwig) {
    Collection c = (Collection)values.remove(PowerForms.getFormValidateName());
    String formid = null;
    if (c != null) {
      for (Iterator i = c.iterator(); i.hasNext(); ) {
	 formid = i.next().toString();
      }
    }
    HtmlForm form = getForm(formid);
    if (form == null) {
      handler.error("invalid form submitted");
      return false;
    } else {
      return form.validate(values, handler, jwig);
    }
  }

  void print(PrintWriter out) {
    /*
    out.println("<!-- forms: "+forms+" -->");
    out.println("<!-- regexps: "+regexps+" -->");
    out.println("<!-- formats: "+formats+" -->");
    */
    if (content != null) {
      content.print(out);
    }
  }

  private String docType = null;

  void setDocType(String docType) {
    this.docType = docType;
  }

  void setContent(HtmlContent content) {
    this.content = content;
    content.setDocument(this);
  }

  public String toString() {
    CharArrayWriter chars = new CharArrayWriter();
    PrintWriter out = new PrintWriter(chars);
    out.println("Document");
    out.println("  Updates (classes): "+constraintClasses);
    out.println("  Elements: "+elements);
    out.println("Forms: ");
    for (int i = 0; i < forms.size(); i++) {
      out.println(forms.get(i));
    }
    out.println("  fields: "+fields);
    out.println("  constraints: "+constraints);
    out.println("  formConstraints: "+formConstraints);
    out.close();
    return chars.toString();
  }

  private static final void DEBUG(Object o) {
    /* System.err.println(String.valueOf(o)); */
  }
}
